/*
 * Copyright (c) 2012, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * */

/*
 *  ======== main.c ========
 *  Example showing use of LogSnapshot APIs and EventHdr2 macros to allow re-assembling memory blocks from snapshot event data.
 */



#define LINUX TRUE

#include <xdc/std.h>
#include <ti/uiaplus/loggers/stream/LoggerStreamer.h>
#include <ti/uia/runtime/LogSnapshot.h>
#include <ti/uia/runtime/UIAPacket.h>
#include <ti/uia/runtime/EventHdr.h>
#include <xdc/runtime/Log.h>
#include <xdc/runtime/Diags.h>
#include <ti/uia/runtime/LogUC.h>

#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>
#include <time.h>


#ifdef LINUX
/* for syscall to getpid */
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#endif


/* Set PRINT_PACKET_INFO to TRUE to print a summary of the snapshot event data in each UIA packet */
#define PRINT_PACKET_INFO (FALSE)

/*
 * Define how much memory each snapshot is required to capture.  Max is 0xFFFF bytes.
 */
#define SNAPSHOT_SIZE_IN_BYTES 50000
#define DATABLOCK_LEN_IN_WORDS (((SNAPSHOT_SIZE_IN_BYTES%4) == 0) ? (SNAPSHOT_SIZE_IN_BYTES >> 2) : ((SNAPSHOT_SIZE_IN_BYTES >> 2)+1))
#define NUM_DATA_BLOCKS 2
/*
 * Define how many UIA packets are available to transfer two snapshot events.
 * For 1024 byte buffers, need 160 packets for 100000 bytes of payload data.
 */
#define LOGGERSTREAMER_MAXEVENTSIZE (256) 
#define LOGGERSTREAMER_BUFSIZE (1024)
#define NUM_SNAPSHOTS_PER_PACKET ((LOGGERSTREAMER_BUFSIZE - 64) / LOGGERSTREAMER_MAXEVENTSIZE)
/* Since snapshot events can be variable length up to the MAXEVENTSIZE, it is a bit tricky to determine
 * how many packets will be generated in order to capture all of the snapshot data.  
 * The following formula has been found to do a good job for buffer sizes of 1024 and greater.  
 * (You may need to hardcode this value if you find that not enough packets are being generated.)
 * In a real system, however, you would not use a static array of packets but would instead use a 
 * pool of packets with full packets being transmitted and returned as empty packets
 * to the pool. 
 */
#define NUM_UIAPACKETS (8+(SNAPSHOT_SIZE_IN_BYTES*2/(NUM_SNAPSHOTS_PER_PACKET*(LOGGERSTREAMER_MAXEVENTSIZE - EventHdr_sizeOfEventHeader_EventWithSnapshotIdAndTimestamp))))
#define BITS32_TO_MAU(bits32)       ((bits32) * sizeof(Bits32))

/* The following array provides the buffers used to store the UIAPackets.  In a real system, these would be transmitted to another CPU or host */
Char buffer[NUM_UIAPACKETS][LOGGERSTREAMER_BUFSIZE];

/*
 * Data to be logged using snapshot events is stored in the following two arrays
 */

Int32 dataBlock1[DATABLOCK_LEN_IN_WORDS];
Int32 dataBlock2[DATABLOCK_LEN_IN_WORDS];

/*
 * The data extracted from the snapshot event payloads will be stored in the following array
 */
Int32 reconstructedDataBlock[NUM_DATA_BLOCKS][DATABLOCK_LEN_IN_WORDS];


/* Some variables used by the extractSnapshotData function */
Int recoveredDataBlockIndex = 0;
Int lastUsedRecoveredDataIndex = 0;
Int lastRecoveredSnapshotId = 0;

/* Some statistics that are collected by the extractSnapshotData funciton */
Int numIncompleteBlocks = 0;
Int numBytesRemainingInCurrentBlock = 0;



volatile int gCount = 0;

void init(Ptr pCtxInfo);
void printSnapshotInfo(uint32_t* pPacketBuffer);
void extractSnapshotData(uint32_t* pPacketBuffer);

int main(int argc, char *argv[])
{
	Int i;
	int numErrors = 0;
        int numBytes = 0;

	// initialize the buffers used to store the event info
	init(NULL);

	for (i=0;i<DATABLOCK_LEN_IN_WORDS;i++){
		dataBlock1[i] = i;
		dataBlock2[i] = (i+DATABLOCK_LEN_IN_WORDS);
		reconstructedDataBlock[0][i] = 0;
		reconstructedDataBlock[1][i] = 0;
	}
	Log_printUC0(Diags_USER1, "Starting test...");

	/* Demo of how to log a large block of memory.
	 * The LogSnapshot_writeMemoryBlock API is called twice, first to log the 50,000 bytes of data in the dataBlock1 array,
	 * and then to log the 50,000 bytes of data in the datablock2 array.
	 * The API calls the LoggerStreamer logger's writeMemoryRange API, which generates a number of snapshot events, each
	 * of which can be up to LOGGERSTREAMER_MAXEVENTSIZE bytes in size.  Each event carries the snapshot ID that is
	 * passed in as the first parameter to the LogSnapshot_writeMemoryBlock API.
	 *
	 * Each snapshot event that is generated carries the following data:
	 * - a timestamp showing when the event was logged
	 * - the snapshot ID that was passed in as the first parameter to the LogSnapshot_writeMemoryBlock API
	 * - the length of the original data that the snapshot was taken of (in this case, 50,000 bytes)
	 * - the number of bytes of data that the event itself is carrying
	 * - a data payload, carrying the above number of bytes of the data that the snapshot was taken of.
	 *
	 * There are also some additional fields (fmt string, data address, etc.) which are provided for informational purposes
	 * so that, when the event is viewed in System Analyzer, users can understand what data the event was intended to capture.
	 * (These informational fields are not required in order to reconstruct the data on a different CPU)
	 *
	 * See the extractSnapshotData function for information on how the snapshot events are decoded and the data payloads
	 * extracted and re-assembled into a copy of the original data array.
	 *
	 * The first parameter to the LogSnapshot_writeMemoryBlock API should be the result returned by
         * the LogSnapshot_getSnapshotId() API, which returns a unique non-zero ID to use to group a set of snapshot event logs together
         * This allows tooling to treat a set of consecutive event logs as a unit and display all of the relevant data together as a set.
	 * The snapshot ID is incremented by one each time the API is called, wrapping around at 0x7FFF back to 1.
         */
	LogSnapshot_writeMemoryBlock(LogSnapshot_getSnapshotId(), "Datablock 1", (IArg)&dataBlock1, DATABLOCK_LEN_IN_WORDS*sizeof(Int32));
        printf("\nAfter first call to LogSnapshot, number of packets generated = %d.\n",gCount);
	LogSnapshot_writeMemoryBlock(LogSnapshot_getSnapshotId(), "Datablock 2", (IArg)&dataBlock2, DATABLOCK_LEN_IN_WORDS*sizeof(Int32));

	// ensure all packets have been given back to the exchangeBuffer function.
	LoggerStreamer_flush();

	Log_printUC1(Diags_USER1, "Ending test.  %d packets generated.",gCount);
	printf("\nAfter second call to LogSnapshot, number of packets generated = %d.\n (number allocated = %d).  UIAPacket size = %d.\n",gCount,NUM_UIAPACKETS,LOGGERSTREAMER_BUFSIZE);

	// iterate through all of the UIAPackets that have been generated
	for (i=0;i<gCount;i++){

#if PRINT_PACKET_INFO
		printf("\n----Packet %ld----\n",UIAPacket_getSequenceCount((UIAPacket_Hdr*)buffer[i]));
		printSnapshotInfo((uint32_t*)buffer[i]);
#endif
		extractSnapshotData((uint32_t*)buffer[i]);
	}

	/* The following check verifies that the reconstructed copy of the original data blocks exactly matches
	 * the original data.  Note that if the endianness of sender was different than endianness of receiver,
	 * you would have needed to swizzle the data in the block after it had been collected.
	 * You can get the endianness of the sender from the UIA packet using the UIAPacket_getPayloadEndianness macro if required.
	 */
	for (i=0;i<DATABLOCK_LEN_IN_WORDS;i++){
		if ((dataBlock1[i] != reconstructedDataBlock[0][i]) || (dataBlock2[i] != reconstructedDataBlock[1][i])){
			numErrors++;
		} else {
                        numBytes += 8;
                }
	}
	printf("\nVerification results:\n number of errors found in %d bytes in reconstructed data blocks = %d\n number of incomplete blocks = %d\n", numBytes,numErrors,numIncompleteBlocks); 
    return 0;
}

UInt disablePreemption(){
	return (0);
}
/*
 * ======== restorePreemption ========
 * Restores the preemption enable state to the
 * value indicated by the key that is passed in as a parameter.
 */
void restorePreemption(UInt key){

}

/*
 * ======== getTimestamp ========
 * This function is called by LoggerStreamer when an event is being logged.
 * It returns the 64b timestamp to use for the event in the two words
 * pointed to by the lsw and msw UInt32* pointers.
 * In order to allow correlation of events from multiple cores on Keystone 
 * architecture devices, the application should be compiled with a predefined
 * symbol (GLOBAL_TIMESTAMP_ADRS) that specifies the virtual address of the 
 * emulation performance counter in the PLLctrl module 
 */
void getTimestamp(UInt32* lsw, UInt32* msw){
#ifdef GLOBAL_TIMESTAMP_ADRS

    *lsw = *(Uint32 *)(GLOBAL_TIMESTAMP_ADRS);
    *msw = *(Uint32 *)(4+GLOBAL_TIMESTAMP_ADRS);

#else
#ifdef LINUX
    struct timeval nowtime;
    unsigned long timestamp;

    gettimeofday(&nowtime,NULL);  // time elapsed since 
    timestamp = (1000000*nowtime.tv_sec)+nowtime.tv_usec;

#else
    unsigned long timestamp = clock();
#endif


    *lsw = (Bits32)(timestamp & 0xFFFFFFFF);  
#if xdc_target__sizeof_Long > xdc_target__sizeof_Int32
    *msw = (Bits32)(timestamp >> 32);  
#else
    *msw = 0;
#endif


#endif


}


/* Called when LoggerStream is full.  Returns the next buffer to write events to, and
 * sends out the just filled event packet over UDP to System Analyzer. */
Ptr bufferExchange(Char *full)
{
	gCount++;
	if (gCount >= NUM_UIAPACKETS){
          gCount = NUM_UIAPACKETS;
	}
	return ((Ptr*)buffer[gCount]);
}


/*
 * initialize the BUF_WITH_CONTEXT data structures, setting a pointer
 * to whatever context information you want to be able to reference when
 * working with the buffers
 */
void init(Ptr pCtxInfo){
	int i;
	short cpuId = 0;
	
    /* Each application needs to generate a CRC16 of the application name so that
     * System Analyzer can identify which executable file to read symbols from in order
     * to decode string constants for the application properly.  When multiple applications
     * are generating events simultaneously, the appName MUST match
     * the name of the executable (without extension or path) in order for System Analyzer
     * to be able to locate the correct file containing each application's symbols.
     * In System Analyzer's UIA Config file, a separate endpoint must be configured 
     * to identify each Linux user-space application that can log events.
     */
	unsigned char const appName[] = "demoWithAdrs";
	short crcApp16 = LoggerStreamer_generateCRC(appName,strlen((char *)appName));
        printf("CRC (16b CCITT): 0x%x.\n",crcApp16);
	printf("Stack: &appName[0] = 0x%x\n",(int)&appName[0]);
	printf("Initialized Data: &gCount = 0x%x\n",(int)&gCount);

	for (i=0;i<NUM_UIAPACKETS;i++){
            LoggerStreamer_initBuffer(buffer[i],cpuId,crcApp16);
	}

	// initialize LoggerStreamer
	LoggerStreamer_init((UInt32*)buffer[0],LOGGERSTREAMER_BUFSIZE);

}


/* ======== extractSnapshotData ========
 * The extractSnapshotData function shows how to
 *  - get the following info from a UIA packet:  packet type, length of a UIA event packet, size of the UIA packet header.
 *  - find the first event in the UIA packet
 *  - iterate through all of the events in a UIA packet
 *  - for each event, determine whether the event is a snapshot event
 *  The snapshot ID is used to associate events that carry parts of the same data set.
 *
 * param uint32_t* pPacketBuffer - pointer to the UIAPacket data that is to be decoded.
 *
 */
void extractSnapshotData(uint32_t* pPacketBuffer){
	uint32_t packetLength = 0;
	uint32_t eventLength = 0;
	uint32_t* pEvent = 0;
        uint32_t* pSnapshotDataPayload = 0;
	uint32_t eventWord1 = 0;
	UInt32 snapshotId;
	int dataSizeInBytes;
	UIAPacket_Hdr* hdr = (UIAPacket_Hdr*)pPacketBuffer;

	/* 1. Check packet header */
	if (UIAPacket_getHdrType(hdr) == UIAPacket_HdrType_EventPkt){
		packetLength = UIAPacket_getEventLength(hdr);

		/* get a pointer to the first event in the packet */
		pEvent = (uint32_t* )((uint32_t)pPacketBuffer + sizeof(UIAPacket_Hdr));
		packetLength -= sizeof(UIAPacket_Hdr);
		/* iterate through the packet, finding all events */
		while (packetLength > 0){
			eventWord1 = *pEvent;
			eventLength = EventHdr_getLength(eventWord1);
			packetLength -= eventLength;

			if (EventHdr_isASnapshotEvent(pEvent)){
				/*  In this example, snapshot ID = 1 is used to identify events that are carrying the data logged by the first
				 *  call of the LogSnapshot_writeMemoryBlock API, and snapshot ID = 2 is used to identify events that are carrying
				 *  the data logged by the second call of the LogSnapshot_writeMemoryBlock API
				 */
				snapshotId = EventHdr_getSnapshotId(pEvent);

				if (lastRecoveredSnapshotId == 0){
					/* this is the first snaphot event that has been received.
					 * save the snapshot ID so that any future events with this snapshot ID
					 * can have their payload stored in the same data array.
					 */
					lastRecoveredSnapshotId = snapshotId;
					/* store how many bytes in total the original call to LogSnapshot_writeMemoryBlock logged */
					numBytesRemainingInCurrentBlock = EventHdr_getSnapshotTotalLength(pEvent);
				} else if (snapshotId != lastRecoveredSnapshotId) {
					/* A new snaphot ID has been encountered.  Start storing the data
					 * from snapshot events with this new snaphot ID in the next buffer in recontstructedDataBlock.
					 */
					lastRecoveredSnapshotId = snapshotId;
					lastUsedRecoveredDataIndex = 0;
					recoveredDataBlockIndex++;

					/* Verify that all of the data for the previous snapshot ID has been received */
					if (numBytesRemainingInCurrentBlock > 0){
						// In a real application, this might occur if the UDP packets were reordered or if a packet was dropped.
						// If there is confidence that packet reordering has not occurred, the snapshot data that has been
						// collected for this snapshotId would be thrown away as incomplete.
						// This simple test application assumes all packets are in order and none are missing.
						numIncompleteBlocks++;
					}
					/* store how many bytes in total the original call to LogSnapshot_writeMemoryBlock logged */
					numBytesRemainingInCurrentBlock = EventHdr_getSnapshotTotalLength(pEvent);
				}
				/* get the data payload size for this event */
				dataSizeInBytes = EventHdr_getSnapshotEventDataLength(pEvent);
				/* update the number of bytes remaining to receive */
				numBytesRemainingInCurrentBlock -= dataSizeInBytes;
				/* copy the payload data into the current reconstructedDataBlock array */
                                pSnapshotDataPayload = (void*)EventHdr_getSnapshotData(pEvent);
                                if (pSnapshotDataPayload != NULL){
                                    memcpy((void*)&reconstructedDataBlock[recoveredDataBlockIndex][lastUsedRecoveredDataIndex], (void*)pSnapshotDataPayload, dataSizeInBytes);
                                }
				/* update the index into the reconstructedDataBlock array, converting from a byte count to an Int32 count by shifting right by 2. */
				lastUsedRecoveredDataIndex += (dataSizeInBytes >> 2);
			}
			/* update the pEvent pointer to point to the next event in the packet */
			pEvent = (uint32_t*)((uint32_t)pEvent + eventLength);
		}
	}
}

/* ======== printSnapshotInfo ========
 * Prints out information about the events in the UIAPacket
 * param uint32_t* pPacketBuffer - pointer to the UIAPacket data that is to be decoded.
 */
void printSnapshotInfo(uint32_t* pPacketBuffer){
	uint32_t packetLength = 0;
	uint32_t eventLength = 0;
	uint32_t* pEvent = 0;
	uint32_t eventWord1 = 0;
	uint32_t* pDataWord1 = NULL;
	uint32_t dataWord1 = 0;
	int num = 1;
	int i=0;
	UIAPacket_Hdr* hdr = (UIAPacket_Hdr*)pPacketBuffer;
	printf("\n----Packet %ld----\n",UIAPacket_getSequenceCount(hdr));
	/* 1. Check packet header */
	if (UIAPacket_getHdrType(hdr) == UIAPacket_HdrType_EventPkt){
		packetLength = UIAPacket_getEventLength(hdr);

		pEvent = (uint32_t* )((uint32_t)pPacketBuffer + sizeof(UIAPacket_Hdr));
		packetLength -= sizeof(UIAPacket_Hdr);
		while (packetLength > 0){
			eventWord1 = *pEvent;
			eventLength = EventHdr_getLength(eventWord1);
			packetLength -= eventLength;

			if (EventHdr_isASnapshotEvent(pEvent)){
				printf("Event %d IS a snapshot event.\n",num);
				for (i=0;i<10;i++){
					printf(" word %d = 0x%x\n",i,*(pEvent+i));
				}
				printf("  Snapshot ID = %ld\n",EventHdr_getSnapshotId(pEvent));
				pDataWord1 = (uint32_t *)EventHdr_getSnapshotData(pEvent);
				printf("  Address of snapshot data carried by the event is 0x%lx\n",(UInt32)pDataWord1);
				dataWord1 = *pDataWord1;
				printf("  First word of snapshot data is 0x%x\n",dataWord1);
				printf("  Snapshot format string = %s\n",(char*)EventHdr_getSnapshotFmtStringAdrs(pEvent));
				printf("  original adrs of snapshot data = 0x%lx\n",EventHdr_getSnapshotDataAdrs(pEvent));
				printf("  num bytes of snapshot data in this event = %ld\n",EventHdr_getSnapshotEventDataLength(pEvent));
				printf("  total num bytes of snapshot data = %ld\n",EventHdr_getSnapshotTotalLength(pEvent));
			} else {
				printf("Event %d is not a snapshot event.\n",num);
			}
			pEvent = (uint32_t *)((uint32_t )pEvent + eventLength);
			num++;
		}
	}
}


