/*
 * Copyright (c) 2012, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
310 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * */
/*
 * ======== cUIA Example Program : Logging Events using LoggerStreamer ========
 * This tutorial demonstrates the use of the Log module to
 * log a number of different types of events.  Please see the 
 * the <cUIA install folder>/docs/cUIA_Linux_Example_Program_User_Guide.html
 * file for more info.
 */

/* Replace the IP address in the following line with the IP address of the PC running System Analyzer */
#define SYSTEM_ANALYZER_PC_IP "158.218.99.52"

#define LINUX TRUE

#include <xdc/std.h>
#include <ti/uiaplus/loggers/stream/LoggerStreamer.h>
#include <ti/uia/runtime/LogSnapshot.h>
#include <ti/uia/events/UIABenchmark.h>
#include <ti/uia/events/UIAEvt.h>
#include <ti/uia/events/UIAErr.h>
#include <ti/uia/events/UIAStatistic.h>
#include <xdc/runtime/Log.h>
#include <xdc/runtime/Diags.h>
#include <ti/uia/runtime/LogUC.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <string.h>
#include <time.h>

#ifdef LINUX
/* for syscall to getpid */
#include <sys/types.h>
#include <unistd.h>
#endif

#define BUFLEN 1024
#define NPACK 10
#define PORT 1235

/* The following constants used as event and error constants
 * for use with the UIAErr and UIAEvt events.
 */


typedef struct _BUF_WITH_CONTEXT {
	Char udpHdr[44]; // note that header is 42 bytes long, and must start at udpHdr[2] OR the data structures need to be aligned
	// so that eventPacket is word aligned.
	UInt32 eventPacket[BUFLEN/sizeof(UInt32)];  // UIA event data packet - this is what the logger works with.  Note that LoggerStreamer.bufSize = 1024;
	Ptr next;
	Ptr pStartOfUDPPacket;  // points to the start of the UDP packet - i.e. what needs to be transmitted over the Ethernet link
	Ptr pContextInfo;   // points to whatever context info you want to associate with this buffer
	Int coreNumber;     // DNUM value for the core that logged the events
} BUF_WITH_CONTEXT;

volatile Int count = 0;
#define NUM_BUFFERS (NPACK)
BUF_WITH_CONTEXT buffer[NUM_BUFFERS];

#define DETAIL_EVENT_CODE (0x1111)
#define CRITICAL_EVENT_CODE (0xCCCC)
#define FATAL_EVENT_CODE (0xFFFF)

int skt;
struct sockaddr_in si_other;

void init(Ptr pCtxInfo);


int main(int argc, char *argv[])
{
	int i=0;
	const struct timespec sleepSetting = {0,100000000};
	struct timespec remSleepSetting;
	char myString[256];
	int snapshotId = 0;
	int loopCtr = 0;
	int pid = 0;
    int param1 = 1234;
    int param2 = 0x12345678;
    int data = 0x7654321;
    int instanceId = 1;
    int numWordsProcessed = 0xABCDEF;
    char destAdrs[128];

    if (argc != 2) {
         /* argc should be 2 for correct execution */ 	
        printf("\nusage: %s 123.45.67.89 \n(replace 123.45.67.89 with the IP address of the PC that System Analyzer is running on.)\n",argv[0]);
    } else {
        strcpy(destAdrs,argv[1]);
	printf("CLOCKS_PER_SEC = %ld\n", CLOCKS_PER_SEC);
        printf("Sending packets to %s\n...\n",destAdrs);
       
#ifdef LINUX
	pid = getpid();
#endif
	// initialize the buffers used to store the event info
	init(NULL);
	skt=socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);

	sprintf(myString,"12345678 testing writeMemoryRange!\n");
	memset((char *) &si_other, 0, sizeof(si_other));
	si_other.sin_family = AF_INET;
	si_other.sin_port = htons(PORT);
	if (inet_aton(destAdrs, &si_other.sin_addr)==0) {
		fprintf(stderr, "inet_aton() failed\n");
		exit(1);
	}



	/* log the events.  Sends out a packet each time the buffer fills up
	 * and the buffer exchange function is called. 
	 */
	while (TRUE){
		/* The Diags_setMask API allows sets of events to be dynamically enabled and disabled */
		if ((loopCtr & 1) == 0) {
			Diags_setMask("Main+FZ");
			Log_warning0("Enabling Info events(+F) and Analysis events (+Z)");
		} else {
			Diags_setMask("Main-FZ");
			Log_warning0("Disabling Info events(-F) and Analysis events (-Z)");
		}
		loopCtr++;
		
        /*  The Log_info APIs (Log_info0..Log_info5) allow you to logging generic
         *  "informational" events that report what line of code logged the event.
         *  They allow you to log a printf-style string along with 0 to 5 values
         *  containing information such as event codes, details about what occurred
         *  application state information, etc.
         */
        Log_info0("Hello World!");

        /* Use Log_warning[0..6] to log an event that reports a warning-level event. */
        Log_warning1("This is a warning with one parameter: %d",param1);
        /* Use Log_error[0..6] to log an event to report that a recoverable error has occurred */
        Log_error2("This is an error with two parameters: %d,0x%x",param1,param2);

        /* The Log_write[0]..[7] APIs provide a lower-level API that allows you to
         * specify a specific event to log.  Unlike the Log_info, Log_warning and Log_error
         * APIs, they do not by default log information about the line of code that
         * the event was logged from.  Each event has a predefined set of parameters that
         * should be logged with the event.  Please read the CDOC documentation for 
         * the event to see what parameters should be logged and for examples that demonstrate
         * the use of the event.
         *
         * For example, the UIAErr module provides a number of predefined events that you can
         * use to log error conditions such as exceptions, critical errors,
         * and fatal errors.  The CDOC documentation for the ti.uia.events.UIAErr module
         * provides a list of all of the error events defined in this module.
         */
        Log_write4(UIAErr_criticalWithStr,CRITICAL_EVENT_CODE,
                        (IArg)"This is what a critical event looks like. It was logged from %$F",
                        (IArg)__FILE__,(IArg)__LINE__);
        Log_write4(UIAErr_fatalWithStr,FATAL_EVENT_CODE,
                        (IArg)"This is what a critical event looks like. It was logged from %$F",
                        (IArg)__FILE__,(IArg)__LINE__);
                        
        /* The UIAEvt module provides a number of predefined events that you can
         * use to log informational events.  The 'detail' level events are lower
         * in priority than info events.
         */
        Log_write5(UIAEvt_detailWithStr,DETAIL_EVENT_CODE,
                        (IArg)"We\'ve just logged a detailed event! data=%d logged from file %s at line %d",
                        data,(IArg)__FILE__,(IArg)__LINE__);

        /* The UIAStatistic module provides predefined events for logging statistical
         * information that can be viewed using System Analyzer's graphs, charts, etc.
         * Each predefined event has a predefined format string so you don't have
         * to specify one as a parameter for the Log_write event.  You will need to
         * check the documentation for the event to see what parameters the
         * format string expects.   If you hover the mouse over the UIAStatistic
         * event in the following line of code, you should see a tooltip that contains
         * a brief description of the event.
         */
        Log_write4(UIAStatistic_wordsProcessedByInstance,
                        (IArg)__FILE__,(IArg)__LINE__,instanceId,numWordsProcessed);

		/* Demo of how to log a dynamically allocated string */
		LogSnapshot_writeString(snapshotId,"test writeString",myString,strlen(myString));

		/* The ti.uia.runtime.Log_writeUCX unconditional write macros do not test the
		 * Diags masks and so take less time to execute, but cannot be dynamically disabled. */
		Log_writeUC1(LoggerStreamer_L_test, 0x800000 | i++);
		Log_writeUC2(LoggerStreamer_L_test, 0x800000 | i++,1);
		Log_writeUC3(LoggerStreamer_L_test, 0x800000 | i++,2,1);
		Log_writeUC4(LoggerStreamer_L_test, 0x800000 | i++,3,2,1);
		Log_writeUC5(LoggerStreamer_L_test, 0x800000 | i++,4,3,2,1);
		Log_writeUC6(LoggerStreamer_L_test, 0x800000 | i++,5,4,3,2,1);
		Log_writeUC7(LoggerStreamer_L_test, 0x800000 | i++,6,5,4,3,2,1);
		Log_writeUC8(LoggerStreamer_L_test, 0x800000 | i++,7,6,5,4,3,2,1);

		/* Examples of the use of ti.uia.runtime.Log_errorUCX unconditional error macros */
		Log_errorUC0("ErrorUC0");
		Log_errorUC1("ErrorUC1",1);
		Log_errorUC2("ErrorUC2",2,1);
		Log_errorUC3("ErrorUC3",3,2,1);
		Log_errorUC4("ErrorUC4",4,3,2,1);
		Log_errorUC5("ErrorUC5",5,4,3,2,1);

		/* Examples of the ti.uia.runtime.Log_warningUCX unconditional warning macros */
		Log_warningUC0("WarningUC0");
		Log_warningUC1("WarningUC1",1);
		Log_warningUC2("WarningUC2",2,1);
		Log_warningUC3("WarningUC3",3,2,1);
		Log_warningUC4("WarningUC4",4,3,2,1);
		Log_warningUC5("WarningUC5",5,4,3,2,1);

		/* Examplse of the ti.uia.runtime.Log_infoUCX unconditional info macros */
		Log_infoUC0("InfoUC0");
		Log_infoUC1("InfoUC1",1);
		Log_infoUC2("InfoUC2",2,1);
		Log_infoUC3("InfoUC3",3,2,1);
		Log_infoUC4("InfoUC4",4,3,2,1);
		Log_infoUC5("InfoUC5",5,4,3,2,1);

		/* Example of the use of the ti.uia.events.UIABenchmark analysis events, used for measuring 
		 * the execution time of a region of code. */
		Log_write1(UIABenchmark_start, (IArg)"benchmark1");
		Log_write1(UIABenchmark_stop, (IArg)"benchmark1");
		Log_write2(UIABenchmark_startInstance,(IArg)"benchmark2: PID=0x%x",pid);
		Log_write2(UIABenchmark_stopInstance,(IArg)"benchmark2: PID=0x%x",pid);

		/* Examples of various other Log APIs that are provided */
		Log_write3(LoggerStreamer_L_test, 0x777000 | i++,2,1);
		Log_write4(LoggerStreamer_L_test, 0x777000 | i++,3,2,1);
		Log_write5(LoggerStreamer_L_test, 0x777000 | i++,4,3,2,1);
		Log_write6(LoggerStreamer_L_test, 0x777000 | i++,5,4,3,2,1);
		Log_write7(LoggerStreamer_L_test, 0x777000 | i++,6,5,4,3,2,1);
		Log_write8(LoggerStreamer_L_test, 0x777000 | i++,7,6,5,4,3,2,1);
		Log_error0("Error 0");
		Log_error1("Error 1",1);
		Log_error2("Error 2",2,1);
		Log_error3("Error 3",3,2,1);
		Log_error4("Error 4",4,3,2,1);
		Log_error5("Error 5",5,4,3,2,1);
		Log_warning0("Warning0");
		Log_warning1("Warning1",1);
		Log_warning2("Warning2",2,1);
		Log_warning3("Warning3",3,2,1);
		Log_warning4("Warning4",4,3,2,1);
		Log_warning5("Warning5",5,4,3,2,1);
		Log_info0("Info0");
		Log_info1("Info1",1);
		Log_info2("Info2",2,1);
		Log_info3("Info3",3,2,1);
		Log_info4("Info4",4,3,2,1);
		Log_info5("Info5",5,4,3,2,1);
		Log_printUC0(xdc_runtime_Diags_ANALYSIS,"Print 0");
		Log_printUC1(xdc_runtime_Diags_ANALYSIS,"Print 1",1);
		/* wait for a bit to avoid generating too many UDP packets! */
		nanosleep(&sleepSetting,&remSleepSetting);
	}
	close(skt);
    }
    return 0;
}

UInt disablePreemption(){
	return (0);
}
/*
 * ======== restorePreemption ========
 * Restores the preemption enable state to the
 * value indicated by the key that is passed in as a parameter.
 */
void restorePreemption(UInt key){

}

/*
 * ======== getTimestamp ========
 * This function is called by LoggerStreamer when an event is being logged.
 * It returns the 64b timestamp to use for the event in the two words
 * pointed to by the lsw and msw UInt32* pointers.
 * In order to allow correlation of events from multiple cores on Keystone 
 * architecture devices, the application should be compiled with a predefined
 * symbol (GLOBAL_TIMESTAMP_ADRS) that specifies the virtual address of the 
 * emulation performance counter in the PLLctrl module 
 */
void getTimestamp(UInt32* lsw, UInt32* msw){
#ifdef GLOBAL_TIMESTAMP_ADRS
    *lsw = *(Uint32 *)(GLOBAL_TIMESTAMP_ADRS);
    *msw = *(Uint32 *)(4+GLOBAL_TIMESTAMP_ADRS);

#else
    Long res = clock();

    *lsw = (Bits32)(res & 0xFFFFFFFF);  // for multicore correlation, use 
#if xdc_target__sizeof_Long > xdc_target__sizeof_Int32
    *msw = (Bits32)(res >> 32);  // for multicore correlation, use 
#else
    *msw = 0;
#endif
#endif

}


/* Called when LoggerStream is full.  Returns the next buffer to write events to, and
 * sends out the just filled event packet over UDP to System Analyzer. */
Ptr bufferExchange(Char *full)
{
	size_t slen=sizeof(si_other);
	int errorCode = 0;
	BUF_WITH_CONTEXT* bufWithCtx = (BUF_WITH_CONTEXT *)((long)full - (long)offsetof(BUF_WITH_CONTEXT,eventPacket));
	BUF_WITH_CONTEXT* nextBufWithCtx = (BUF_WITH_CONTEXT *)bufWithCtx->next;
	const struct sockaddr* pSockAddr = (struct sockaddr *)&si_other;

	printf("Sending packet %d\n", count++);

	errorCode = sendto(skt, full, (size_t)BUFLEN, (int) 0, pSockAddr, slen);
	if (errorCode == -1){
		printf("ERROR: bufferExchange call to sendto returned %d.\n",errorCode);
	}

	return ((Ptr)&nextBufWithCtx->eventPacket);
}



/*
 * initialize the BUF_WITH_CONTEXT data structures, setting a pointer
 * to whatever context information you want to be able to reference when
 * working with the buffers
 */
void init(Ptr pCtxInfo){
	int i;
	short cpuId = 0;

    /* Each application needs to generate a CRC16 of the application name so that
     * System Analyzer can identify which executable file to read symbols from in order
     * to decode string constants for the application properly.  When multiple applications
     * are generating events simultaneously, the appName MUST match
     * the name of the executable (without extension or path) in order for System Analyzer
     * to be able to locate the correct file containing each application's symbols.
     * In System Analyzer's UIA Config file, a separate endpoint must be configured 
     * to identify each Linux user-space application that can log events.
     */
	unsigned char const appName[] = "demoWithAdrs";
	short crcApp16 = LoggerStreamer_generateCRC(appName,strlen((char *)appName));
        printf("CRC (16b CCITT): 0x%x.\n",crcApp16);
	printf("Stack: &appName[0] = 0x%x\n",(int)&appName[0]);

	for (i=0;i<NUM_BUFFERS;i++){
		if (i < (NUM_BUFFERS-1)){
			buffer[i].next = &buffer[i+1];
		} else {
			buffer[i].next = &buffer[0];
		}

		buffer[i].pContextInfo = pCtxInfo;
		buffer[i].pStartOfUDPPacket = &buffer[i].udpHdr[2];
		buffer[i].coreNumber = cpuId;
		LoggerStreamer_initBuffer(buffer[i].eventPacket,cpuId,crcApp16);
	}

	// initialize LoggerStreamer
	LoggerStreamer_init(buffer[0].eventPacket,BUFLEN);

}


