/*
 * Copyright (c) 2012, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * */

#ifndef LOGGERSTREAMER_SUPPORT_H_
#define LOGGERSTREAMER_SUPPORT_H_

/* The following functions must be implemented by the application code that uses LoggerStreamer */

/*
 * ======== disablePreemption ========
 * If multiple threads can preemptively call into the
 * Logger, then this function should disable preemption
 * until the restorePreemption function is called.
 * Returns an integer value that can be used to ensure
 * that the original preemption enable state of the
 * application is restored properly.
 */
extern UInt disablePreemption();
/*
 * ======== restorePreemption ========
 * Restores the preemption enable state to the
 * value indicated by the key that is passed in as a parameter.
 */
extern void restorePreemption(UInt key);
/*
 * ======== getTimestamp ========
 * Assigns a timestamp value to the event as it is being written.
 * The 32 LSBs of the timestamp are to be written into the
 * address pointed to by lsw, and the 32 MSBs of the timestamp
 * are to be written into the address pointed to by msw.
 * e.g.
 * 	*lsw = *(Uint32 *)(GLOBAL_TIMESTAMP_ADRS);
 *  *msw = *(Uint32 *)(4+GLOBAL_TIMESTAMP_ADRS);
 */
extern void getTimestamp(UInt32* lsw, UInt32* msw);

/*
 * ======== bufferExchange ========
 *  LoggerStreamer stores events in a buffer provided by the application
 *  and will call this application provided exchange function when it is full.
 *  The exchange is called within the context of a Log call, so the exchange
 *  function should be design to fast. Since the exchange function is called
 *  within the context of the Log call, LoggerStreamer guarantees no Log
 *  records are dropped (i.e. LoggerStreamer is lossless).
 *
 *  The application is responsible for providing the buffers that LoggerStreamer
 *  uses. These buffers are to be initialized by calling the LoggerStreamer_initBuffer
 *  macro (defined in LoggerStreamer__epilogue.h).  The
 *
 *  The buffer that is provided is used to the Log records. The first
 *  four 32-bit words contains a `{@link UIAPacket#Hdr}` structure. This struct
 *  is used by the host (e.g. System Analyzer in CCS) to help decode the
 *  data (e.g. endianess, length of data, etc.). The `{@link UIAPacket#Hdr}`
 *  structure is initialized via the `{@link #initBuffer}` API. All buffers
 *  given to LoggerStreamer (via priming or exchange) must be initialized via
 *  `{@link #initBuffer}`.
 *
 *  The size of buffer -*- includes UIAPacket_Hdr?
 *  LoggerStreamer treats the buffer as an UInt32 array. So the application
 *  must guarantee the buffers are aligned accordingly.
 *
 *  When the buffer is filled, LoggerStreamer will hand it off to the
 *  application using the exchange function (`{@link #exchangeFxn}`).
 *  The exchange function must be of type `{@link #ExchangeFxnType}`.
 *  The exchange function is called in the context of a log
 *  so generally the exchange function should be quick to execute.
 *
 *  Interrupts are disabled during the duration of the log call including
 *  when the exchange function is called. LoggerStreamer will ignore any
 *  log events generated during the exchangeFxn (e.g. posting a semaphore).
 *
 *
 */
extern xdc_Ptr bufferExchange(Ptr buffer);


#endif /* LOGGERSTREAMER_SUPPORT_H_ */
